import torch
import cv2
import numpy as np
from pathlib import Path
from ultralytics import YOLO  # YOLOv8 model

# Přidání cesty k SORT trackeru (relativní cesta)
import sys
sys.path.append(str(Path(__file__).parent / "sort"))
from sort import Sort

# Nastavení relativních cest
current_dir = Path(__file__).parent
video_path = current_dir / "data" / "cyklo_sestrih.mp4"
output_path = current_dir / "outputs" / "output.mp4"

# Načtení modelu YOLOv8
model = YOLO('yolov8s.pt') 

# Načtení videa
cap = cv2.VideoCapture(str(video_path))

if not cap.isOpened():
    print("Chyba při otevření videa.")
    exit()

# Inicializace SORT trackeru
tracker = Sort()

# Zjistíme rozměry videa
frame_width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
frame_height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
fps = cap.get(cv2.CAP_PROP_FPS)
line_position = frame_width // 2  # Vertikální čára ve středu videa

# Inicializace počítadel pro směr cyklistů
count_left_to_right = 0
count_right_to_left = 0
tracked_paths = {}  # Pro sledování trajektorií cyklistů

# Funkce pro ověření, zda objekt je v rámci offsetu kolem čáry
def is_within_offset(line_position, center_x, offset):
    return (line_position - offset <= center_x <= line_position + offset)

# Nastavení VideoWriter pro ukládání výstupu
fourcc = cv2.VideoWriter_fourcc(*'mp4v')
out = cv2.VideoWriter(str(output_path), fourcc, fps, (frame_width, frame_height))

if not out.isOpened():
    print("Chyba při otevření VideoWriter.")
    cap.release()
    exit()

while cap.isOpened():
    ret, frame = cap.read()
    if not ret:
        break

    # Převod obrázku na RGB (OpenCV používá BGR)
    img_rgb = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
    
    # Detekce objektů pomocí YOLO
    results = model(img_rgb)

    # Příprava detekcí pro cyklisty
    detections = []
    for detection in results:
        for box in detection.boxes:
            x1, y1, x2, y2 = box.xyxy[0].tolist()
            conf = box.conf[0]
            class_id = box.cls[0]
            if conf > 0.5 and int(class_id) == 1:  # Cyklista (class_id == 1)
                detections.append([x1, y1, x2, y2, conf.item()])

    # Konverze detekcí na numpy pole pro vstup do SORT
    detections_np = np.array(detections)
    
    # Aktualizace trackeru s aktuálními detekcemi
    if detections_np.size > 0:
        tracked_objects = tracker.update(detections_np)
    else:
        tracked_objects = np.empty((0, 5))  # Zajistit, že tracked_objects má správný formát

    # Zobrazení vertikální čáry
    cv2.line(frame, (line_position, 0), (line_position, frame_height), (255, 0, 0), 2)

    # Zpracování sledovaných cyklistů
    for obj in tracked_objects:
        x1, y1, x2, y2, track_id = obj
        center_x = int((x1 + x2) / 2)

        # Sledování trajektorie cyklisty
        if track_id not in tracked_paths:
            tracked_paths[track_id] = [center_x]
        else:
            tracked_paths[track_id].append(center_x)
        
        # Kontrola směru pohybu
        path = tracked_paths[track_id]
        if len(path) > 2:
            # Pohyb zleva doprava
            if path[-2] < line_position <= path[-1] and is_within_offset(line_position, center_x, 25):
                count_left_to_right += 1
                tracked_paths.pop(track_id)  # Odstraníme sledovanou cestu, aby se cyklista nezapočítal vícekrát

            # Pohyb zprava doleva
            elif path[-2] > line_position >= path[-1] and is_within_offset(line_position, center_x, 25):
                count_right_to_left += 1
                tracked_paths.pop(track_id)

        # Nakreslení rámečku kolem detekovaného cyklisty a přidání textu s confidence hodnotou
        cv2.rectangle(frame, (int(x1), int(y1)), (int(x2), int(y2)), (0, 255, 0), 2)
        confidence_text = f"Cyclist: {conf:.2f}"
        cv2.putText(frame, confidence_text, (int(x1), int(y1) - 10), cv2.FONT_HERSHEY_SIMPLEX, 0.9, (0, 255, 0), 2)

    # Zobrazení počtu cyklistů podle směru
    cv2.putText(frame, f'Left to Right: {count_left_to_right}', (50, 50), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 255, 0), 2)
    cv2.putText(frame, f'Right to Left: {count_right_to_left}', (50, 100), cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 255, 0), 2)
    
    # Uložení snímku do nového videa
    out.write(frame)

cap.release()
out.release()
cv2.destroyAllWindows()

# Vytisknutí konečných výsledků
print(f'Final Left to Right Count: {count_left_to_right}')
print(f'Final Right to Left Count: {count_right_to_left}')
