import arcpy
import os
import math
import os
arcpy.env.parallelProcessingFactor = "100%"  # Nebo nastavte konkrétní počet jader, např. "4"

def main():
    try:
        # Parametry (vstupy a výstupy)
        rides_layer = arcpy.GetParameterAsText(0)  # Vstupní vrstva jízd
        osm_layer = arcpy.GetParameterAsText(1)  # Vstupní vrstva OSM
        buffer_distance = arcpy.GetParameterAsText(2)  # Šířka bufferu (např. "10 Meters")

        # Získání cesty k geodatabázi vstupní vrstvy OSM
        input_gdb = os.path.dirname(osm_layer)
        print("Krok 1: Generování bodů podél linií...")

        # Výstupní názvy vrstev v geodatabázi (uložíme je do stejné geodatabáze jako vstupní OSM)
        points_output = f"{input_gdb}/osm_points"  # Body podél linií
        split_lines_output = f"{input_gdb}/osm_split_lines"  # Rozdělené linie
        buffer_output = f"{input_gdb}/osm_buffer"
        intersect_output = f"{input_gdb}/intersect_output"
        filtered_output = f"{input_gdb}/filtered_lines"
        summary_table = f"{input_gdb}/summary_table"

        # Krok 1: Vytvoření bodů podél linií (každých 100 m)
      
        split_lines_output=osm_layer
            # Krok 3: Přidání pole s délkou každého úseku (v metrech)
        print("Krok 3: Přidání pole s délkou každého úseku...")
        arcpy.AddMessage("Krok 3: Přidání pole s délkou každého úseku...")
        arcpy.AddField_management(split_lines_output, "Length_m", "DOUBLE")
        arcpy.CalculateField_management(
            split_lines_output, "Length_m", "!SHAPE.LENGTH@METERS!", "PYTHON3"
        )
        print(f"Délka byla vypočítána pro každý úsek.")
        arcpy.AddMessage(f"Délka byla vypočítána pro každý úsek.")

        # Krok 4: Přidání geometrických atributů (počáteční a koncové body)
        print("Krok 4: Přidání geometrických atributů (počáteční a koncové body)...")
        arcpy.AddMessage("Krok 4: Přidání geometrických atributů (počáteční a koncové body)...")
        try:
            # Přidání nových polí, pokud ještě neexistují
            arcpy.management.AddField(split_lines_output, "start_x", "DOUBLE")
            arcpy.management.AddField(split_lines_output, "start_y", "DOUBLE")
            arcpy.management.AddField(split_lines_output, "end_x", "DOUBLE")
            arcpy.management.AddField(split_lines_output, "end_y", "DOUBLE")

            # Použití kurzoru pro výpočet geometrických atributů
            with arcpy.da.UpdateCursor(split_lines_output, ["SHAPE@", "start_x", "start_y", "end_x", "end_y"]) as cursor:
                count = 0
                for row in cursor:
                    # Získání počátečního a koncového bodu linie
                    start_point = row[0].firstPoint
                    end_point = row[0].lastPoint

                    # Aktualizace hodnot v polích
                    row[1] = start_point.X
                    row[2] = start_point.Y
                    row[3] = end_point.X
                    row[4] = end_point.Y

                    # Uložení změn do aktuálního řádku
                    cursor.updateRow(row)

                    # Výpis každých 100 000 záznamů pro sledování průběhu
                    count += 1
                    if count % 100000 == 0:
                        print(f"Zpracováno {count} řádků...")
                        arcpy.AddMessage(f"Zpracováno {count} řádků...")

            print(f"Geometrické atributy byly přidány do vrstvy: {split_lines_output}")
            arcpy.AddMessage(f"Geometrické atributy byly přidány do vrstvy: {split_lines_output}")

        except Exception as e:
            print(f"Chyba při přidávání geometrických atributů: {e}")
            arcpy.AddError(f"Chyba při přidávání geometrických atributů: {e}")

        # Krok 5: Výpočet azimutu pro každý úsek linie
        print("Krok 5: Výpočet azimutu pro každý úsek linie...")
        arcpy.AddMessage("Krok 5: Výpočet azimutu pro každý úsek linie...")

        def calculate_consistent_azimuth(start_x, start_y, end_x, end_y):
            """
            Výpočet konzistentního azimutu mezi dvěma body.
            """
            azimuth = (90 - math.degrees(math.atan2((end_y - start_y), (end_x - start_x)))) % 360
            return azimuth

        try:
            # Přidání pole pro azimut
            if "Azimuth" not in [f.name for f in arcpy.ListFields(split_lines_output)]:
                arcpy.management.AddField(split_lines_output, "Azimuth", "DOUBLE")

            # Výpočet azimutu pomocí kurzoru
            with arcpy.da.UpdateCursor(split_lines_output, ["start_x", "start_y", "end_x", "end_y", "Azimuth"]) as cursor:
                for row in cursor:
                    start_x, start_y = row[0], row[1]
                    end_x, end_y = row[2], row[3]
                    row[4] = calculate_consistent_azimuth(start_x, start_y, end_x, end_y)
                    cursor.updateRow(row)

            print("Azimuty byly úspěšně vypočítány.")
            arcpy.AddMessage("Azimuty byly úspěšně vypočítány.")
        except Exception as e:
            print(f"Chyba při výpočtu azimutů: {e}")
            arcpy.AddError(f"Chyba při výpočtu azimutů: {e}")



        # Přidání nového pole 'id_spoj' a výpočet z OBJECTID
        print("Přidání pole 'id_spoj' a výpočet hodnot z OBJECTID...")
        arcpy.AddMessage("Přidání pole 'id_spoj' a výpočet hodnot z OBJECTID...")
        arcpy.AddField_management(split_lines_output, "id_spoj", "LONG")  # Přidání nového pole typu LONG
        arcpy.CalculateField_management(
            split_lines_output, "id_spoj", "!OBJECTID_1!", "PYTHON3"  # Výpočet hodnoty z OBJECTID
        )
        print(f"Pole 'id_spoj' bylo přidáno a vypočítáno.")
        arcpy.AddMessage(f"Pole 'id_spoj' bylo přidáno a vypočítáno.")

        # Krok 4: Vytvoření bufferu kolem rozdělených linií
        print("Krok 4: Vytvoření bufferu kolem rozdělených linií...")
        arcpy.AddMessage("Krok 4: Vytvoření bufferu kolem rozdělených linií...")
        arcpy.analysis.Buffer(
        split_lines_output, 
        buffer_output, 
        buffer_distance, 
        dissolve_option="NONE", 
    )
        print(f"Buffer byl vytvořen: {buffer_output}")
        arcpy.AddMessage(f"Buffer byl vytvořen: {buffer_output}")

    # Použití PairwiseIntersect namísto Intersect
        arcpy.PairwiseIntersect_analysis([rides_layer, buffer_output], intersect_output)
        
        print(f"Průsečíky byly vytvořeny pomocí Pairwise Intersect: {intersect_output}")
        arcpy.AddMessage(f"Průsečíky byly vytvořeny pomocí Pairwise Intersect: {intersect_output}")
        try:
            # Přidání nových polí, pokud ještě neexistují
            arcpy.management.AddField(intersect_output, "start_x", "DOUBLE")
            arcpy.management.AddField(intersect_output, "start_y", "DOUBLE")
            arcpy.management.AddField(intersect_output, "end_x", "DOUBLE")
            arcpy.management.AddField(intersect_output, "end_y", "DOUBLE")
            arcpy.management.AddField(intersect_output, "delka_intersect", "DOUBLE")  # Pole pro délku linie
            arcpy.management.AddField(intersect_output, "LengthRatio", "DOUBLE")  # Pole pro poměr délky

            # Použití kurzoru pro výpočet geometrických atributů a délky
            with arcpy.da.UpdateCursor(intersect_output, ["SHAPE@", "start_x", "start_y", "end_x", "end_y", "delka_intersect"]) as cursor:
                count = 0
                for row in cursor:
                    # Získání počátečního a koncového bodu linie
                    start_point = row[0].firstPoint
                    end_point = row[0].lastPoint

                    # Aktualizace hodnot v polích pro počáteční a koncové body
                    row[1] = start_point.X
                    row[2] = start_point.Y
                    row[3] = end_point.X
                    row[4] = end_point.Y

                    # Výpočet délky linie v metrech
                    length_meters = row[0].length  # Vrací délku v jednotkách prostorového referenčního systému
                    row[5] = length_meters  # Uložení délky do pole 'delka_intersect'

                    # Výpočet poměru délky (příklad: poměr vůči celkové délce 1000 m)
                 
                    # Uložení změn do aktuálního řádku
                    cursor.updateRow(row)

                    # Zvýšení počítadla a výpis každých 100 000 záznamů pro sledování průběhu
                    count += 1
                    if count % 100000 == 0:
                        print(f"Zpracováno {count} řádků...")
                        arcpy.AddMessage(f"Zpracováno {count} řádků...")

            print(f"Geometrické atributy a délka byly přidány do vrstvy: {intersect_output}")
            arcpy.AddMessage(f"Geometrické atributy a délka byly přidány do vrstvy: {intersect_output}")

        except Exception as e:
            print(f"Chyba při přidávání geometrických atributů a délky: {e}")
            arcpy.AddError(f"Chyba při přidávání geometrických atributů a délky: {e}")

    

        # Výpočet azimutu pro každý úsek linie
        print("Výpočet azimutu pro každý úsek linie v intersect_output...")
        arcpy.AddMessage("Výpočet azimutu pro každý úsek linie v intersect_output...")

        def calculate_consistent_azimuth(start_x, start_y, end_x, end_y):
            """
            Výpočet konzistentního azimutu mezi dvěma body.
            """
            azimuth = (90 - math.degrees(math.atan2((end_y - start_y), (end_x - start_x)))) % 360
            return azimuth

        try:
            # Přidání pole pro azimut
            if "Azimuth_intersect" not in [f.name for f in arcpy.ListFields(intersect_output)]:
                arcpy.management.AddField(intersect_output, "Azimuth_intersect", "DOUBLE")

            # Výpočet azimutu pomocí kurzoru
            with arcpy.da.UpdateCursor(intersect_output, ["start_x", "start_y", "end_x", "end_y", "Azimuth_intersect"]) as cursor:
                for row in cursor:
                    start_x, start_y = row[0], row[1]
                    end_x, end_y = row[2], row[3]
                    row[4] = calculate_consistent_azimuth(start_x, start_y, end_x, end_y)
                    cursor.updateRow(row)

            print("Azimuty byly úspěšně vypočítány.")
            arcpy.AddMessage("Azimuty byly úspěšně vypočítány.")
        except Exception as e:
            print(f"Chyba při výpočtu azimutů: {e}")
            arcpy.AddError(f"Chyba při výpočtu azimutů: {e}")
   
        # Použití UpdateCursor pro výpočet hodnoty 'LengthRatio'
        print("Výpočet pole 'LengthRatio' pomocí UpdateCursor...")
        arcpy.AddMessage("Výpočet pole 'LengthRatio' pomocí UpdateCursor...")

        # Definice polí pro UpdateCursor
        fields = ["delka_intersect", "Length_m", "LengthRatio"]  # delka_intersect, Length_m a LengthRatio jsou pole

        # Počítadlo pro sledování zpracovaných řádků
        row_count = 0
        total_rows = int(arcpy.GetCount_management(intersect_output)[0])  # Získání celkového počtu řádků

        with arcpy.da.UpdateCursor(intersect_output, fields) as cursor:
            for row in cursor:
                # Výpočet poměru délky (příklad: delka_intersect dělená (Length_m + 20))
                row[2] = row[0] / (row[1] + 20)  # row[0] = delka_intersect, row[1] = Length_m, row[2] = LengthRatio
                cursor.updateRow(row)

                # Zvýšení počítadla a výpis každých 10 000 záznamů
                row_count += 1
                if row_count % 10000 == 0:
                    print(f"Zpracováno {row_count} záznamů z {total_rows}...")
                    arcpy.AddMessage(f"Zpracováno {row_count} záznamů z {total_rows}...")

        print("Výpočet pole 'LengthRatio' byl dokončen.")
        arcpy.AddMessage("Výpočet pole 'LengthRatio' byl dokončen.")


        # Finální zpráva
        print("Výpočet poměru délky dokončen.")
        arcpy.AddMessage("Výpočet poměru délky dokončen.")







        # Krok 6: Výpočet rozdílu azimutů
        print("Krok 6: Výpočet rozdílu azimutů...")
        arcpy.AddMessage("Krok 6: Výpočet rozdílu azimutů...")

        def calculate_azimuth_difference(input_lines):
            """
            Výpočet rozdílu azimutů mezi dvěma poli (Azimuth a Azimuth_1).
            """
            try:
                # Přidání pole pro rozdíl azimutů
                if "difference_azimuth" not in [f.name for f in arcpy.ListFields(input_lines)]:
                    arcpy.management.AddField(input_lines, "difference_azimuth", "DOUBLE")

                # Iterace přes záznamy pomocí kurzoru
                with arcpy.da.UpdateCursor(input_lines, ["Azimuth", "Azimuth_intersect", "difference_azimuth"]) as cursor:
                    for row in cursor:
                        azimuth1 = row[0]
                        azimuth2 = row[1]

                        # Kontrola na None
                        if azimuth1 is None or azimuth2 is None:
                            row[2] = None  # Nebo nastavte výchozí hodnotu
                        else:
                            difference = abs(azimuth1 - azimuth2)
                            row[2] = min(difference, 180 - difference)  # Rozdíl v rozmezí 0-180 stupňů

                        cursor.updateRow(row)

                arcpy.AddMessage("Rozdíl azimutů byl úspěšně vypočítán.")
            except Exception as e:
                arcpy.AddError(f"Chyba při výpočtu rozdílu azimutů: {e}")

        # Vypočítání rozdílu azimutů
        calculate_azimuth_difference(intersect_output)

        # Krok 7: Filtrování linií podle délky a rozdílu azimutů
        print("Krok 7: Filtrování linií podle délky a rozdílu azimutů...")
        arcpy.AddMessage("Krok 7: Filtrování linií podle délky a rozdílu azimutů...")

        try:
            # Vytvoření dočasné vrstvy pro filtrování
            arcpy.MakeFeatureLayer_management(intersect_output, "temp_layer")

            # Definice SQL dotazu pro filtraci
            query = "(delka > 15 AND LengthRatio >= 0.7 AND LengthRatio <= 1.3)OR(delka <= 15 AND LengthRatio >= 0.7 AND LengthRatio <= 1.3 AND difference_azimuth < 40)"
            
            # Filtrování na základě podmínek
            arcpy.SelectLayerByAttribute_management("temp_layer", "NEW_SELECTION", query)

            # Uložení filtrovaných řádků do nové vrstvy
            filtered_output = os.path.join(os.path.dirname(intersect_output), "filtered_intersect_output")
            arcpy.CopyFeatures_management("temp_layer", filtered_output)

            print(f"Řádky byly úspěšně filtrovány a uloženy do vrstvy: {filtered_output}")
            arcpy.AddMessage(f"Řádky byly úspěšně filtrovány a uloženy do vrstvy: {filtered_output}")
        except Exception as e:
            print(f"Chyba při filtrování řádků: {e}")
            arcpy.AddError(f"Chyba při filtrování řádků: {e}")





        arcpy.CopyFeatures_management("temp_layer", filtered_output)
        print(f"Linie byly filtrovány a uloženy do vrstvy: {filtered_output}")
        arcpy.AddMessage(f"Linie byly filtrovány a uloženy do vrstvy: {filtered_output}")

            # Krok 8: Sumarizace počtu jízd na úsek OSM podle atributu 'rok'
        print("Krok 8: Sumarizace počtu jízd na úsek OSM podle atributu 'rok'...")
        arcpy.AddMessage("Krok 8: Sumarizace počtu jízd na úsek OSM podle atributu 'rok'...")

        # Vytvoření sumarizační tabulky
        arcpy.Statistics_analysis(
            filtered_output, 
            summary_table, 
            [["OBJECTID", "COUNT"]], 
            ["id_spoj", "rok"]
        )
        print(f"Sumarizace byla dokončena. Tabulka výsledků je uložena zde: {summary_table}")
        arcpy.AddMessage(f"Sumarizace byla dokončena. Tabulka výsledků je uložena zde: {summary_table}")

        # Krok 9: Pivotování tabulky podle roku
        print("Krok 9: Pivotování tabulky podle roku...")
        arcpy.AddMessage("Krok 9: Pivotování tabulky podle roku...")

        # Výstupní pivotovaná tabulka
        pivot_table = f"{input_gdb}/pivot_summary"
        arcpy.PivotTable_management(
            summary_table, 
            "id_spoj", 
            "rok", 
            "FREQUENCY", 
            pivot_table
        )
        print(f"Pivotovaná tabulka byla vytvořena a uložena zde: {pivot_table}")
        arcpy.AddMessage(f"Pivotovaná tabulka byla vytvořena a uložena zde: {pivot_table}")

        # Krok 10: Připojení výsledků zpět k původním úsekům OSM
        print("Krok 10: Připojení výsledků zpět k původním úsekům OSM...")
        arcpy.AddMessage("Krok 10: Připojení výsledků zpět k původním úsekům OSM...")

        # Výstupní vrstva s počty jízd
        output_osm_with_counts = f"{input_gdb}/osm_with_counts"
        arcpy.JoinField_management(split_lines_output, "id_spoj", pivot_table, "id_spoj")
        arcpy.CopyFeatures_management(split_lines_output, output_osm_with_counts)

        print(f"Výsledky byly připojeny zpět a uloženy do vrstvy: {output_osm_with_counts}")
        arcpy.AddMessage(f"Výsledky byly připojeny zpět a uloženy do vrstvy: {output_osm_with_counts}")

    except Exception as e:
        # Zpracování chyb
        print(f"Došlo k chybě: {e}")
        arcpy.AddError(f"Došlo k chybě: {e}")

if __name__ == "__main__":
    main()
