import arcpy
import math
import os

def get_workspace_from_path(output_path):
    """
    Získání workspace (databáze nebo složky) z cesty k výstupní vrstvě.
    """
    return os.path.dirname(output_path)

def split_lines(input_line, output_points, interval, output_lines):
    """
    Rozdělení linií na úseky podle zvoleného intervalu.
    """
    try:
        interval_str = f"{interval} Meters"

        # Vytvoření bodů podél linie
        arcpy.management.GeneratePointsAlongLines(
            input_line, output_points, "DISTANCE", interval_str
        )
        arcpy.AddMessage("Body podél linie byly úspěšně vytvořeny.")

        # Rozdělení linie podle bodů
        arcpy.management.SplitLineAtPoint(input_line, output_points, output_lines, "1 Meters")
        arcpy.AddMessage("Linie byla úspěšně rozdělena podle bodů.")

    except Exception as e:
        arcpy.AddError(f"Chyba při rozdělení linií: {e}")

def add_geometry_attributes(input_layer):
    """
    Přidání geometrických atributů (X, Y startu a cíle) do rozdělených linií pomocí CalculateGeometryAttributes.
    """
    try:
        arcpy.management.CalculateGeometryAttributes(
            in_features=input_layer,
            geometry_property="start_x LINE_START_X;start_y LINE_START_Y;end_x LINE_END_X;end_y LINE_END_Y",
            length_unit="",
            area_unit="",
            coordinate_system=None,
            coordinate_format="SAME_AS_INPUT"
        )
        arcpy.AddMessage(f"Geometrické atributy byly přidány do vrstvy: {input_layer}")
    except Exception as e:
        arcpy.AddError(f"Chyba při přidávání geometrických atributů: {e}")




def calculate_consistent_azimuth(start_x, start_y, end_x, end_y):
        """
        Výpočet konzistentního azimutu mezi dvěma body.
        """
        azimuth = (90 - math.degrees(math.atan2((end_y - start_y), (end_x - start_x)))) % 360
        if azimuth > 180:
            azimuth -= 180
        return azimuth
   







def calculate_azimuth_for_lines(input_lines, azimuth_field="Azimuth"):
    """
    Výpočet azimutu pro každý úsek linie a uložení do specifikovaného pole.
    
    :param input_lines: Název vstupní vrstvy (feature class nebo shapefile).
    :param azimuth_field: Název pole, kam se uloží vypočítaný azimut (výchozí: "Azimuth").
    """
    try:
        # Přidání pole pro azimut, pokud ještě neexistuje
        if azimuth_field not in [f.name for f in arcpy.ListFields(input_lines)]:
            arcpy.management.AddField(input_lines, azimuth_field, "DOUBLE")

        # Výpočet azimutu pro každý úsek
        with arcpy.da.UpdateCursor(input_lines, ["start_x", "start_y", "end_x", "end_y", azimuth_field]) as cursor:
            for row in cursor:
                start_x, start_y = row[0], row[1]
                end_x, end_y = row[2], row[3]
                # Funkce pro výpočet azimutu
                row[4] = calculate_consistent_azimuth(start_x, start_y, end_x, end_y)
                cursor.updateRow(row)

        arcpy.AddMessage(f"Azimuty byly úspěšně vypočítány a uloženy do pole '{azimuth_field}'.")
    except Exception as e:
        arcpy.AddError(f"Chyba při výpočtu azimutů: {e}")




def create_centroids(input_lines, output_centroids):
    """
    Vytvoření středových bodů pro rozsekané úseky OSM linie.
    """
    try:
        # Vytvoření centroidů pro každý úsek linie
        arcpy.management.FeatureToPoint(input_lines, output_centroids, "CENTROID")
        arcpy.AddMessage(f"Středové body byly vytvořeny: {output_centroids}")
    except Exception as e:
        arcpy.AddError(f"Chyba při vytváření středových bodů: {e}")

def spatial_join_nearest(input_points, target_lines, max_distance, output_joined):
    """
    Vyhledání nejbližšího úseku z cílové vrstvy a připojení jeho atributů ke středovým bodům.
    """
    try:
        # Prostorové spojení s nejbližším úsekem z cílové vrstvy
        arcpy.analysis.SpatialJoin(
            target_features=input_points,
            join_features=target_lines,
            out_feature_class=output_joined,
            join_type="KEEP_COMMON",
            match_option="CLOSEST",
            search_radius=f"{max_distance} Meters"
        )
        arcpy.AddMessage(f"Nejbližší úseky byly připojeny: {output_joined}")
    except Exception as e:
        arcpy.AddError(f"Chyba při prostorovém spojení: {e}")

def join_attributes_back_to_lines(source_points_with_attributes, split_lines_osm):
    """
    Připojení atributů z bodů zpět k rozděleným liniím OSM.
    """
    try:
        # Připojení atributů zpět k rozděleným liniím na základě ORIG_FID

        fieldList = ["id", "prac_2016", "vik_2016","vik_2018", "prac_2018", "prac_2020","vik_2020","prac_2022", "vik_2022", "Azimuth_cyklo"]
        arcpy.management.JoinField(
            in_data=split_lines_osm,
            in_field="OBJECTID_1",
            join_table=source_points_with_attributes,
            join_field="TARGET_FID",
            fields=fieldList)
        arcpy.AddMessage("Atributy byly připojeny zpět k rozděleným liniím.")
    except Exception as e:
        arcpy.AddError(f"Chyba při připojování atributů zpět k liniím: {e}")

def join_attributes_back_to_lines_car(source_points_with_attributes, split_lines_osm):
    """
    Připojení atributů z bodů zpět k rozděleným liniím OSM.
    """
    try:
        # Připojení atributů zpět k rozděleným liniím na základě ORIG_FID

        fieldList = ["id", "f2023_car", "f2023_truc","f2022_car", "f2022_truc",  "Azimuth_auto"]
        arcpy.management.JoinField(
            in_data=split_lines_osm,
            in_field="OBJECTID_1",
            join_table=source_points_with_attributes,
            join_field="TARGET_FID",
            fields=fieldList)
        arcpy.AddMessage("Atributy byly připojeny zpět k rozděleným liniím auta.")
    except Exception as e:
        arcpy.AddError(f"Chyba při připojování atributů zpět k liniím auta: {e}")



def add_geometry_attributes2(input_layer):
    """
    Přidání geometrických atributů (X, Y startu a cíle) do rozdělených linií pomocí CalculateGeometryAttributes.
    """
    try:
        # Převod délky na metry pomocí geodetické metody
        arcpy.management.AddField(input_layer, "Shape_Length_M", "DOUBLE")
        
        arcpy.management.CalculateGeometryAttributes(
        in_features=input_layer,
        geometry_property="Shape_Length_M LENGTH_GEODESIC",
        length_unit="METERS",
        area_unit="",
        coordinate_system=None,
        coordinate_format="SAME_AS_INPUT"
)
        
        arcpy.AddMessage(f"Délky v metrech byly přidány do vrstvy: {input_layer}")
    except Exception as e:
        arcpy.AddError(f"Chyba při přidávání geometrických atributů: {e}")


def azimuth_difference(input_lines, azimuth_field_1, azimuth_field_2, field_name="difference_azimuth"):
    """
    Vypočítá rozdíl mezi dvěma azimuty a uloží jej do specifikovaného pole.
    
    :param input_lines: Název vstupní vrstvy (feature class nebo shapefile).
    :param azimuth_field_1: Název prvního pole s azimutem (např. Azimuth_cyklo).
    :param azimuth_field_2: Název druhého pole s azimutem (např. Azimuth_auto).
    :param field_name: Název pole, kam se uloží rozdíl azimutů (výchozí: "difference_azimuth").
    """
    try:
        # Přidání pole pro rozdíl azimutů, pokud ještě neexistuje
        if field_name not in [f.name for f in arcpy.ListFields(input_lines)]:
            arcpy.management.AddField(input_lines, field_name, "DOUBLE")

        # Iterace přes záznamy pomocí kurzoru
        with arcpy.da.UpdateCursor(input_lines, [azimuth_field_1, azimuth_field_2, field_name]) as cursor:
            for row in cursor:
                azimuth1 = row[0]
                azimuth2 = row[1]

                # Kontrola na None hodnoty
                if azimuth1 is None or azimuth2 is None:
                    row[2] = None  # Nebo nastavte výchozí hodnotu
                else:
                    difference = abs(azimuth1 - azimuth2)
                    row[2] = min(difference, 180 - difference)  # Kruhový rozdíl

                cursor.updateRow(row)

        arcpy.AddMessage(f"Rozdíl mezi azimuty byl úspěšně vypočítán a uložen do pole '{field_name}'.")
    except Exception as e:
        arcpy.AddError(f"Chyba při výpočtu rozdílu azimutů: {e}")



def calculate_slope(input_lines, dem_raster):
    """
    Výpočet sklonu úseku na základě rozdílu výšek mezi počátečním a koncovým bodem linie.
    """
    try:
        # Workspace pro mezivýstupy
        workspace = get_workspace_from_path(input_lines)
        start_points = os.path.join(workspace, "start_points")
        end_points = os.path.join(workspace, "end_points")

        # Krok 1: Extrakce počátečních a koncových bodů
        arcpy.management.FeatureVerticesToPoints(
            in_features=input_lines,
            out_feature_class=start_points,
            point_location="START"
        )
        arcpy.management.FeatureVerticesToPoints(
            in_features=input_lines,
            out_feature_class=end_points,
            point_location="END"
        )
        arcpy.AddMessage("Počáteční a koncové body byly úspěšně vytvořeny.")

        # Krok 2: Přidání výšek z DEM
        arcpy.ddd.AddSurfaceInformation(
            in_feature_class=start_points,
            in_surface=dem_raster,
            out_property="Z",
            z_factor=1
        )
        arcpy.ddd.AddSurfaceInformation(
            in_feature_class=end_points,
            in_surface=dem_raster,
            out_property="Z",
            z_factor=1
        )
        arcpy.AddMessage("Výšky byly přidány k počátečním a koncovým bodům.")

        # Krok 3: Spojení výšek zpět k liniím
        arcpy.management.JoinField(
            in_data=input_lines,
            in_field="OBJECTID_1",
            join_table=start_points,
            join_field="ORIG_FID",
            fields=["Z"]
        )
        arcpy.management.JoinField(
            in_data=input_lines,
            in_field="OBJECTID_1",
            join_table=end_points,
            join_field="ORIG_FID",
            fields=["Z"]
        )
        arcpy.AddMessage("Výšky byly připojeny zpět k liniím.")

        # Krok 4: Výpočet sklonu (%)
        if "Slope" not in [f.name for f in arcpy.ListFields(input_lines)]:
            arcpy.management.AddField(input_lines, "Slope", "DOUBLE")

        arcpy.management.CalculateField(
            in_table=input_lines,
            field="Slope",
            expression="abs(!Z! - !Z_1!) / !Shape_Length_M! * 100",
            expression_type="PYTHON3"
        )
        arcpy.AddMessage("Sklon byl úspěšně vypočítán.")
    except Exception as e:
        arcpy.AddError(f"Chyba při výpočtu sklonu: {e}")
     





def validate_highway_and_oneway(input_lines):
    """
    Validace vlastností highway a oneway na základě zadaných kritérií.
    """
    try:
        # Přidání pole pro validaci
        if "obousmer" not in [f.name for f in arcpy.ListFields(input_lines)]:
            arcpy.management.AddField(input_lines, "obousmer", "SHORT")

        # Definice platných typů highway
        valid_highways = ['secondary', 'primary', 'tertiary', 'trunk']

        # Iterace přes záznamy a kontrola podmínek
        with arcpy.da.UpdateCursor(input_lines, ["highway", "oneway", "obousmer"]) as cursor:
            for row in cursor:
                highway_type = row[0]
                oneway_status = row[1]

                # Validace: highway musí být v seznamu a oneway musí být 'yes'
                if highway_type in valid_highways and oneway_status == 'yes':
                    row[2] = 1  # Platné
                else:
                    row[2] = 0  # Neplatné

                cursor.updateRow(row)

        arcpy.AddMessage("Validace vlastností highway a oneway byla úspěšně dokončena.")
    except Exception as e:
        arcpy.AddError(f"Chyba při validaci vlastností: {e}")



if __name__ == "__main__":
    # Parametry pro první linii (OSM)
    input_osm = arcpy.GetParameterAsText(0)
    output_points_osm = arcpy.GetParameterAsText(1)
    interval_osm = float(arcpy.GetParameterAsText(2))
    split_lines_osm = arcpy.GetParameterAsText(3)

    # Parametry pro druhou linii (pentlogram)
    input_pentlogram = arcpy.GetParameterAsText(4)
    output_points_pentlogram = arcpy.GetParameterAsText(5)
    interval_pentlogram = float(arcpy.GetParameterAsText(6))
    split_lines_pentlogram = arcpy.GetParameterAsText(7)
    max_distance = float(arcpy.GetParameterAsText(8))  # Maximální vzdálenost v metrech

      # Parametry pro druhou linii (pentlogram_car)
    input_pentlogram_car = arcpy.GetParameterAsText(9)
    output_points_pentlogram_car = arcpy.GetParameterAsText(10)
    interval_pentlogram_car = float(arcpy.GetParameterAsText(11))
    split_lines_pentlogram_car = arcpy.GetParameterAsText(12)
    max_distance_car = float(arcpy.GetParameterAsText(13))  # Maximální vzdálenost v metrech

    # Získání workspace z cesty k výstupním liniím OSM
    workspace = get_workspace_from_path(split_lines_osm)

    # Cesty pro mezivýpočty v rámci stejné databáze/složky jako výstupy
    centroids_osm = os.path.join(workspace, "centroids_osm")
    joined_output = os.path.join(workspace, "joined_output")
    joined_output_car = os.path.join(workspace, "joined_output_car")


    # Parametry pro vyhledávání nejbližších úseků

    dem_raster = arcpy.GetParameterAsText(14)  # DEM raster pro získání výšek
    # Rozdělení OSM linie a pentlogramu
    split_lines(input_osm, output_points_osm, interval_osm, split_lines_osm)
    split_lines(input_pentlogram, output_points_pentlogram, interval_pentlogram, split_lines_pentlogram)
    split_lines(input_pentlogram_car, output_points_pentlogram_car, interval_pentlogram_car, split_lines_pentlogram_car)

    # Přidání geometrických atributů
    add_geometry_attributes(split_lines_osm)
    add_geometry_attributes(split_lines_pentlogram)
    add_geometry_attributes(split_lines_pentlogram_car)

    # Výpočet azimutů
    calculate_azimuth_for_lines(split_lines_osm)


    calculate_azimuth_for_lines(split_lines_pentlogram, azimuth_field="Azimuth_cyklo")

    calculate_azimuth_for_lines(split_lines_pentlogram_car, azimuth_field="Azimuth_auto")

    # Vytvoření středových bodů pro rozsekanou OSM linii
    create_centroids(split_lines_osm, centroids_osm)

    # Prostorové spojení s nejbližšími úseky pentlogramu
    spatial_join_nearest(centroids_osm, split_lines_pentlogram, max_distance, joined_output)
    spatial_join_nearest(centroids_osm, split_lines_pentlogram_car, max_distance_car, joined_output_car)

    # Připojení atributů zpět k rozdělené OSM linii
    join_attributes_back_to_lines(joined_output, split_lines_osm)

    join_attributes_back_to_lines_car(joined_output_car, split_lines_osm)

    add_geometry_attributes2(split_lines_osm)


    #Rozdíl azimutů



    azimuth_difference(split_lines_osm, "Azimuth", "Azimuth_cyklo", field_name="difference_azimuth_cyklo")
    azimuth_difference(split_lines_osm, "Azimuth", "Azimuth_auto", field_name="difference_azimuth_auto")

    # Výpočet sklonu

    calculate_slope(split_lines_osm,dem_raster)

    #výpočet jednosměrnosti

    validate_highway_and_oneway(split_lines_osm)
